"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var jsx_runtime_1 = require("react/jsx-runtime");
var react_1 = require("react");
var MIN_INDEX = 0;
var normalizeValue = function (min, value, max) { return Math.max(Math.min(value, max), min); };
var getPropName = function (axis) { return ({
    minHeight: axis === 'y' ? 'minHeight' : 'minWidth',
    height: axis === 'y' ? 'height' : 'width',
    maxHeight: axis === 'y' ? 'maxHeight' : 'maxWidth',
    top: axis === 'y' ? 'top' : 'left',
    bottom: axis === 'y' ? 'bottom' : 'right',
    scrollHeight: axis === 'y' ? 'scrollHeight' : 'scrollWidth',
    clientHeight: axis === 'y' ? 'clientHeight' : 'clientWidth',
    scrollTop: axis === 'y' ? 'scrollTop' : 'scrollLeft',
    overflowY: axis === 'y' ? 'overflowY' : 'overflowX'
}); };
var getStyleFabric = function (propName) { return function (size) {
    var _a;
    return (_a = {},
        _a[propName.minHeight] = size,
        _a[propName.height] = size,
        _a[propName.maxHeight] = size,
        _a.overflowAnchor = 'none',
        _a.pointerEvents = 'none',
        _a.userSelect = 'none',
        _a.padding = 0,
        _a.margin = 0,
        _a.border = 'none',
        _a);
}; };
var getScrollBottom = function (node, propName) {
    return node[propName.scrollHeight] - node[propName.clientHeight] - node[propName.scrollTop];
};
var setScrollBottom = function (node, value, propName) {
    node[propName.scrollTop] = node[propName.scrollHeight] - node[propName.clientHeight] - value;
};
var setScrollIfNeeded = function (node, propName) {
    if (node) {
        var scrollBottom_1 = getScrollBottom(node, propName);
        setTimeout(function () {
            if (node && getScrollBottom(node, propName) !== scrollBottom_1) {
                setScrollBottom(node, scrollBottom_1, propName);
            }
        });
    }
};
var ViewportList = react_1.forwardRef(function (_a, ref) {
    var viewportRef = _a.viewportRef, _b = _a.items, items = _b === void 0 ? [] : _b, itemMinSize = _a.itemMinSize, _c = _a.margin, margin = _c === void 0 ? 0 : _c, _d = _a.overscan, overscan = _d === void 0 ? 1 : _d, _e = _a.axis, axis = _e === void 0 ? 'y' : _e, _f = _a.initialIndex, initialIndex = _f === void 0 ? 0 : _f, _g = _a.initialAlignToTop, initialAlignToTop = _g === void 0 ? true : _g, children = _a.children;
    var propName = react_1.useMemo(function () { return getPropName(axis); }, [axis]);
    var getStyle = react_1.useMemo(function () { return getStyleFabric(propName); }, [propName]);
    var maxIndex = items.length - 1;
    var itemMinSizeWithMargin = itemMinSize + margin;
    var overscanSize = overscan * itemMinSizeWithMargin;
    var _h = react_1.useState(function () {
        var normalizedInitialIndex = normalizeValue(MIN_INDEX, initialIndex, maxIndex);
        return [normalizedInitialIndex, normalizedInitialIndex];
    }), _j = _h[0], startIndex = _j[0], endIndex = _j[1], setIndexes = _h[1];
    var topRef = react_1.useRef(null);
    var bottomRef = react_1.useRef(null);
    var cache = react_1.useRef([]);
    var scrollToIndex = react_1.useRef(startIndex ? { index: startIndex, alignToTop: initialAlignToTop } : null);
    var step = react_1.useRef(function () { });
    var normalizedStartIndex = normalizeValue(MIN_INDEX, startIndex, maxIndex);
    var normalizedEndIndex = normalizeValue(normalizedStartIndex, endIndex, maxIndex);
    var topStyle = react_1.useMemo(function () {
        return getStyle(cache.current
            .slice(MIN_INDEX, normalizedStartIndex)
            .reduce(function (sum, next) { return sum + next; }, normalizedStartIndex * itemMinSizeWithMargin));
    }, [normalizedStartIndex, itemMinSizeWithMargin, getStyle]);
    var bottomStyle = react_1.useMemo(function () {
        return getStyle(cache.current
            .slice(normalizedEndIndex + 1, maxIndex)
            .reduce(function (sum, next) { return sum + next; }, itemMinSizeWithMargin * (maxIndex - normalizedEndIndex)));
    }, [normalizedEndIndex, maxIndex, itemMinSizeWithMargin, getStyle]);
    var slicedItems = react_1.useMemo(function () {
        return items
            .slice(normalizedStartIndex, normalizedEndIndex + 1)
            .map(function (item, index) { return children(item, normalizedStartIndex + index); });
    }, [items, normalizedStartIndex, normalizedEndIndex, children]);
    step.current = function () {
        if (!viewportRef.current ||
            !topRef.current ||
            !bottomRef.current ||
            !(topRef.current.nextSibling instanceof Element) ||
            !(bottomRef.current.previousSibling instanceof Element)) {
            return;
        }
        var viewportRect = viewportRef.current.getBoundingClientRect();
        var topLimit = viewportRect[propName.top] - overscanSize;
        var bottomLimit = viewportRect[propName.bottom] + overscanSize;
        var topElementRect = topRef.current.nextSibling.getBoundingClientRect();
        var bottomElementRect = bottomRef.current.previousSibling.getBoundingClientRect();
        var maxItemsCountInViewPort = Math.ceil((bottomLimit - topLimit) / itemMinSizeWithMargin);
        var nextStartIndex = startIndex;
        var nextEndIndex = endIndex;
        var diff;
        var index;
        var element;
        if (scrollToIndex.current) {
            var targetIndex = normalizeValue(MIN_INDEX, scrollToIndex.current.index, maxIndex);
            if (targetIndex >= startIndex && targetIndex <= endIndex) {
                index = startIndex;
                element = topRef.current.nextSibling;
                while (element instanceof Element && element !== bottomRef.current) {
                    if (index === targetIndex) {
                        element.scrollIntoView(scrollToIndex.current.alignToTop);
                        scrollToIndex.current = null;
                        break;
                    }
                    element = element.nextSibling;
                    ++index;
                }
                return;
            }
            nextStartIndex = targetIndex - maxItemsCountInViewPort;
            nextEndIndex = targetIndex + maxItemsCountInViewPort;
        }
        else if (topElementRect[propName.top] >= bottomLimit) {
            diff = topElementRect[propName.top] - bottomLimit;
            nextEndIndex = startIndex;
            while (diff >= 0 && nextEndIndex > MIN_INDEX) {
                diff -= (cache.current[--nextEndIndex] || 0) + itemMinSizeWithMargin;
            }
            nextStartIndex = nextEndIndex - maxItemsCountInViewPort;
            setScrollIfNeeded(viewportRef.current, propName);
        }
        else if (bottomElementRect[propName.bottom] + margin <= topLimit) {
            diff = topLimit - bottomElementRect[propName.bottom] + margin;
            nextStartIndex = endIndex;
            while (diff >= 0 && nextStartIndex < maxIndex) {
                diff -= (cache.current[++nextStartIndex] || 0) + itemMinSizeWithMargin;
            }
            nextEndIndex = nextStartIndex + maxItemsCountInViewPort;
        }
        else {
            if (topElementRect[propName.bottom] + margin < topLimit) {
                ++nextStartIndex;
            }
            else if (topElementRect[propName.top] >= topLimit) {
                diff = topElementRect[propName.top] - topLimit;
                while (diff >= 0 && nextStartIndex > MIN_INDEX) {
                    diff -= (cache.current[--nextStartIndex] || 0) + itemMinSizeWithMargin;
                }
                setScrollIfNeeded(viewportRef.current, propName);
            }
            if (bottomElementRect[propName.bottom] + margin <= bottomLimit) {
                diff = bottomLimit - bottomElementRect[propName.bottom] - margin;
                while (diff >= 0 && nextEndIndex < maxIndex) {
                    diff -= (cache.current[++nextEndIndex] || 0) + itemMinSizeWithMargin;
                }
            }
            else if (bottomElementRect[propName.top] > bottomLimit) {
                --nextEndIndex;
            }
        }
        nextStartIndex = normalizeValue(MIN_INDEX, nextStartIndex, maxIndex);
        nextEndIndex = normalizeValue(nextStartIndex, nextEndIndex, maxIndex);
        if (nextStartIndex !== startIndex || nextEndIndex !== endIndex) {
            index = startIndex;
            element = topRef.current.nextSibling;
            while (element instanceof Element && index <= endIndex && element !== bottomRef.current) {
                cache.current[index++] = element[propName.clientHeight] - itemMinSize;
                element = element.nextSibling;
            }
            setIndexes([nextStartIndex, nextEndIndex]);
        }
    };
    react_1.useImperativeHandle(ref, function () { return ({
        scrollToIndex: function (index, alignToTop) {
            if (index === void 0) { index = -1; }
            if (alignToTop === void 0) { alignToTop = true; }
            scrollToIndex.current = { index: index, alignToTop: alignToTop };
        }
    }); }, []);
    react_1.useEffect(function () {
        var frameId;
        var frame = function () {
            step.current();
            frameId = requestAnimationFrame(frame);
        };
        frame();
        return function () {
            cancelAnimationFrame(frameId);
        };
    }, []);
    return (jsx_runtime_1.jsxs(react_1.Fragment, { children: [jsx_runtime_1.jsx("div", { ref: topRef, style: topStyle }, void 0), slicedItems, jsx_runtime_1.jsx("div", { ref: bottomRef, style: bottomStyle }, void 0)] }, void 0));
});
exports.default = ViewportList;
